import abc

import torch
from torch import Tensor


class Instrument(abc.ABC):
    """Base class for all financial instruments."""

    @abc.abstractmethod
    def simulate(self, time_horizon, n_paths=1, init_price=1.0) -> None:
        """Simulate time series of prices of itself (for a primary instrument)
        or its underlier (for a derivative).
        """

    @abc.abstractmethod
    def to(self, *args, **kwargs):
        """Performs dtype and/or device conversion of the time series of prices.

        Args:
            dtype (torch.dtype): Desired floating point type of the floating point
                values of simulated time series.
            device (torch.device): Desired device of the values of simulated time
                series.

        Returns:
            self
        """

    @property
    def dinfo(self) -> list:
        """Returns list of strings that tell `dtype` and `device` of `self`.

        Intended to be used in `__repr__`.

        If `dtype` (`device`) is the one specified in default type,
        `dinfo` will not have the information of it.

        Returns:
            list[str]
        """
        # Implementation here refers to the function `_str_intern` in `pytorch/_tensor_str.py`.

        dinfo = []

        dtype = getattr(self, "dtype", None)
        if dtype is not None:
            if dtype != torch.get_default_dtype():
                dinfo.append("dtype=" + str(dtype))

        # A general logic here is we only print device when it doesn't match
        # the device specified in default tensor type.
        device = getattr(self, "device", None)
        if device is not None:
            if device.type != torch._C._get_default_device() or (
                device.type == "cuda" and torch.cuda.current_device() != device.index
            ):
                dinfo.append("device='" + str(device) + "'")

        return dinfo


class Primary(Instrument):
    """Base class for all primary instruments.

    A primary instrument is a basic financial instrument which is traded on a market
    and therefore the price is accessible as the market price.
    Examples include stocks, bonds, commodities, and currencies.

    Derivatives are issued based on primary instruments
    (See :class:`Derivative` for details).

    Attributes:
        prices (torch.Tensor): The prices of the instrument.
            This attribute is supposed to be set by a method `simulate()`.
            Shape is :math:`(N, T)` where :math:`N` is the number of simulated paths
            and :math:`T` is the number of time steps.
    """

    @abc.abstractmethod
    def simulate(self, time_horizon, n_paths=1, init_price=1.0, **kwargs) -> None:
        """Simulate time series of prices and set an attribute `prices`.

        Args:
            time_horizon (float): The period of time to simulate the price.
            n_paths (int, default=1): The number of paths to simulate.
            init_price (float, default=1.0): The initial value of the prices.
        """

    def to(self, *args, **kwargs):
        """Performs dtype and/or device conversion of the time series of the prices.

        Args:
            dtype (torch.dtype): Desired floating point type of the floating point
                values of simulated time series.
            device (torch.device): Desired device of the values of simulated time
                series.

        Returns:
            self
        """
        device, dtype, *_ = torch._C._nn._parse_to(*args, **kwargs)

        if dtype is not None and not dtype.is_floating_point:
            raise TypeError(
                f"Instrument.to only accepts floating point "
                f"dtypes, but got desired dtype={dtype}"
            )

        if not hasattr(self, "dtype") or dtype is not None:
            self.dtype = dtype
        if not hasattr(self, "device") or device is not None:
            self.device = device

        # If prices have been already simulated, move it
        if hasattr(self, "prices"):
            self.prices = self.prices.to(*args, **kwargs)

        return self


class Derivative(Instrument):
    """Base class for all derivatives.

    A derivative is a financial instrument whose payoff is contingent on
    a primary instrument (or a set of primary instruments).
    A (over-the-counter) derivative is not traded on the market and therefore the price
    is not directly accessible.
    Examples include options and swaps.

    A derivative relies on primary assets (See :class:`Primary` for details), such as
    stocks, bonds, commodities, and currencies.

    Attributes:
        underlier (:class:`Primary`): The underlying asset on which the derivative's
            payoff relies.
    """

    underlier: Primary
    maturity: float

    @property
    def dtype(self) -> torch.dtype:
        return self.underlier.dtype

    @property
    def device(self) -> torch.device:
        return self.underlier.device

    def simulate(self, n_paths=1, init_price=1.0, **kwargs) -> None:
        """
        Simulates time series of the underlier's prices.

        Args:
            n_paths (int): The number of paths to simulate.
            init_price (float): The initial value of the prices.
        """
        self.underlier.simulate(
            time_horizon=self.maturity, n_paths=n_paths, init_price=init_price, **kwargs
        )

    def to(self, *args, **kwargs):
        """Performs dtype and/or device conversion of the time series of the prices.

        Args:
            dtype (torch.dtype): Desired floating point type of the floating point
                values of simulated time series.
            device (torch.device): Desired device of the values of simulated time
                series.

        Returns:
            self
        """
        self.underlier.to(*args, **kwargs)
        return self

    @abc.abstractmethod
    def payoff(self) -> Tensor:
        """Returns the payoffs of the derivative.
        The payoffs is computed based on the prices of `underlier`.

        Returns:
            torch.Tensor
        """



from torch import Tensor

from ..nn.functional import american_binary_payoff
from ._base import Derivative


class AmericanBinaryOption(Derivative):
    """An American binary Option.

    An American binary call option pays an unit amount of cash if and only if
    the maximum of the underlying asset's price until maturity is equal or greater
    than the strike price.

    An American binary put option pays an unit amount of cash if and only if
    the minimum of the underlying asset's price until maturity is equal or smaller
    than the strike price.

    The payoff of an American binary call option is given by:

    .. math ::

        \\mathrm{payoff} =
        \\begin{cases}
            1 & (\\mathrm{Max} \\geq K) \\\\
            0 & (\\text{otherwise})
        \\end{cases}

    Here, :math:`\\mathrm{Max}` is the maximum of the underlying asset's price
    until maturity and :math:`K` is the strike price (`strike`) of the option.

    The payoff of an American binary put option is given by:

    .. math ::

        \\mathrm{payoff} =
        \\begin{cases}
            1 & (\\mathrm{Min} \\leq K) \\\\
            0 & (\\text{otherwise})
        \\end{cases}

    Here, :math:`\\mathrm{Min}` is the minimum of the underlying asset's price.

    Args:
        underlier (:class:`Primary`): The underlying instrument of the option.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.
        maturity (float, default=20/250): The maturity of the option.
        dtype (torch.device, optional): Desired device of returned tensor.
            Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Examples:

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import AmericanBinaryOption
        >>> _ = torch.manual_seed(42)
        >>> deriv = AmericanBinaryOption(BrownianStock(), maturity=5/250, strike=1.01)
        >>> deriv.simulate(n_paths=2)
        >>> deriv.underlier.prices
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])
        >>> deriv.payoff()
        tensor([0., 1.])
    """

    def __init__(
        self,
        underlier,
        call: bool = True,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        dtype=None,
        device=None,
    ):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity
        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")
        params += self.dinfo
        return self.__class__.__name__ + "(" + ", ".join(params) + ")"

    def payoff(self) -> Tensor:
        return american_binary_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



import torch
from torch import Tensor

from pfhedge._utils.bisect import bisect

from ._base import BSModuleMixin


class BSAmericanBinaryOption(BSModuleMixin):
    """Black-Scholes formula for an American Binary Option.

    Args:
        derivative (:class:`pfhedge.instruments.AmericanBinaryOption`, optional):
            The derivative to get the Black-Scholes formula.
        call (bool, default=True): Specifies whether the option is call
            or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - Input: :math:`(N, *, 4)`, where :math:`*` means any number of additional
          dimensions. See `inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`. Delta of the derivative.
          All but the last dimension are the same shape as the input.

    .. seealso ::

        - :class:`pfhedge.nn.BlackScholes`:
          Initialize Black-Scholes formula module from a derivative.

    Examples:

        The `forward` method returns delta of the derivative.

        >>> from pfhedge.nn import BSAmericanBinaryOption
        >>> m = BSAmericanBinaryOption(strike=1.0)
        >>> m.inputs()
        ['log_moneyness', 'max_log_moneyness', 'expiry_time', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, -0.01, 0.1, 0.2],
        ...     [ 0.00,  0.00, 0.1, 0.2],
        ...     [ 0.01,  0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[1.1285],
                [0.0000],
                [0.0000]])

        One can instantiate it using an
        :class:`pfhedge.instruments.AmericanBinaryOption`.

        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import AmericanBinaryOption
        >>> deriv = AmericanBinaryOption(BrownianStock(), strike=1.1)
        >>> m = BSAmericanBinaryOption(deriv)
        >>> m
        BSAmericanBinaryOption(strike=1.1)
    """

    def __init__(self, derivative=None, call: bool = True, strike: float = 1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

    def extra_repr(self):
        params = []
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def delta(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
    ) -> Tensor:
        """Returns delta of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor
        """
        s, m, t, v = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, volatility)
        )

        sqrt2 = torch.sqrt(torch.as_tensor(2.0)).item()
        d1 = self.d1(s, t, v)
        d2 = self.d2(s, t, v)
        c1 = self.N.cdf(d1 / sqrt2)
        p1 = self.N.pdf(d1 / sqrt2)
        p2 = self.N.pdf(d2 / sqrt2)

        d = (1 + c1 + (p1 + p2)) / (2 * self.strike)

        return torch.where(m < 0, d, torch.zeros_like(d))

    @torch.enable_grad()
    def gamma(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
    ) -> Tensor:
        """Returns gamma of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor
        """
        prices = self.strike * torch.exp(torch.as_tensor(log_moneyness))
        prices = Tensor.requires_grad_(prices)

        s = torch.log(prices / self.strike)
        m, t, v = map(torch.as_tensor, (max_log_moneyness, expiry_time, volatility))
        m = torch.max(s, m)  # `s + epsilon` may be greater than `m`

        delta = self.delta(s, m, t, v)
        gamma = torch.autograd.grad(
            delta, prices, grad_outputs=torch.ones_like(prices)
        )[0]

        return gamma

    def price(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
    ) -> Tensor:
        """
        Returns price of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        s, m, t, v = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, volatility)
        )

        sqrt2 = torch.sqrt(torch.as_tensor(2.0)).item()
        n1 = self.N.cdf(self.d1(s, t, v) / sqrt2)
        n2 = self.N.cdf(self.d2(s, t, v) / sqrt2)

        p = (1 / 2) * (torch.exp(s) * (1 + n1) + n2)

        return torch.where(m < 0, p, torch.ones_like(p))

    def implied_volatility(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        price: Tensor,
        precision: float = 1e-6,
    ) -> Tensor:
        """Returns implied volatility of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.
            precision (float, default=1e-6): Computational precision of the implied
                volatility.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            torch.Tensor
        """
        s, m, t, p = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, price)
        )
        get_price = lambda volatility: self.price(s, m, t, volatility)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



import torch
from torch import Tensor

from pfhedge._utils.bisect import bisect

from ._base import BSModuleMixin


class BSLookbackOption(BSModuleMixin):
    """Black-Scholes formula for a lookback option with a fixed strike.

    Args:
        derivative (:class:`pfhedge.instruments.LookbackOption`, optional):
            The derivative to get the Black-Scholes formula.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - Input: :math:`(N, *, 4)`, where :math:`*` means any number of additional
          dimensions.  See `inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`. Delta of the derivative.
          All but the last dimension are the same shape as the input.

    .. seealso ::

        - :class:`pfhedge.nn.BlackScholes`:
          Initialize Black-Scholes formula module from a derivative.

    Examples:

        The `forward` method returns delta of the derivative.

        >>> from pfhedge.nn import BSLookbackOption
        >>> m = BSLookbackOption()
        >>> m.inputs()
        ['log_moneyness', 'max_log_moneyness', 'expiry_time', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, -0.01, 0.1, 0.2],
        ...     [ 0.00,  0.00, 0.1, 0.2],
        ...     [ 0.01,  0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[0.9208],
                [1.0515],
                [1.0515]])

        One can instantiate it using an
        :class:`pfhedge.instruments.LookbackOption`.

        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import LookbackOption
        >>> deriv = LookbackOption(BrownianStock(), strike=1.1)
        >>> m = BSLookbackOption(deriv)
        >>> m
        BSLookbackOption(strike=1.1)
    """

    def __init__(self, derivative=None, call: bool = True, strike: float = 1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

    def extra_repr(self):
        params = []
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def inputs(self) -> list:
        return ["log_moneyness", "max_log_moneyness", "expiry_time", "volatility"]

    @torch.enable_grad()
    def delta(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
        create_graph: bool = False,
    ) -> Tensor:
        """Returns delta of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the prices of the underlying
                asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.
            create_graph (bool, default=False): If True, graph of the derivative
                will be constructed. This option is used to compute gamma.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`


        Returns:
            Tensor
        """
        prices = self.strike * torch.exp(torch.as_tensor(log_moneyness))
        prices = Tensor.requires_grad_(prices)

        s = torch.log(prices / self.strike)
        m, t, v = map(torch.as_tensor, (max_log_moneyness, expiry_time, volatility))
        m = torch.max(s, m)  # `s + epsilon` may be greater than `m`

        price = self.price(s, m, t, v)
        delta = torch.autograd.grad(
            price,
            inputs=prices,
            grad_outputs=torch.ones_like(prices),
            create_graph=create_graph,
        )[0]

        return delta

    @torch.enable_grad()
    def gamma(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
    ) -> Tensor:
        """Returns gamma of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the prices of the
                underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor):
                Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        prices = self.strike * torch.exp(torch.as_tensor(log_moneyness))
        prices = Tensor.requires_grad_(prices)

        s = torch.log(prices / self.strike)
        m, t, v = map(torch.as_tensor, (max_log_moneyness, expiry_time, volatility))
        m = torch.max(s, m)  # `s + epsilon` may be greater than `m`

        delta = self.delta(s, m, t, v, create_graph=True)
        gamma = torch.autograd.grad(
            delta, prices, grad_outputs=torch.ones_like(prices)
        )[0]

        return gamma

    def price(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        volatility: Tensor,
    ) -> Tensor:
        """Returns price of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the prices of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        s, m, t, v = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, volatility)
        )

        d1 = self.d1(s, t, v)
        d2 = self.d2(s, t, v)
        e1 = (s - m + (v ** 2 / 2) * t) / (v * torch.sqrt(t))  # d' in paper
        e2 = (s - m - (v ** 2 / 2) * t) / (v * torch.sqrt(t))

        # when max moneyness < strike
        price_0 = self.strike * (
            torch.exp(s) * self.N.cdf(d1)
            - self.N.cdf(d2)
            + torch.exp(s) * v * torch.sqrt(t) * (d1 * self.N.cdf(d1) + self.N.pdf(d1))
        )
        # when max moneyness >= strike
        price_1 = self.strike * (
            torch.exp(s) * self.N.cdf(e1)
            - torch.exp(m) * self.N.cdf(e2)
            + torch.exp(m)
            - 1
            + torch.exp(s) * v * torch.sqrt(t) * (e1 * self.N.cdf(e1) + self.N.pdf(e1))
        )

        return torch.where(m < 0, price_0, price_1)

    def implied_volatility(
        self,
        log_moneyness: Tensor,
        max_log_moneyness: Tensor,
        expiry_time: Tensor,
        price: Tensor,
        precision: float = 1e-6,
    ) -> Tensor:
        """Returns implied volatility of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the prices of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            price (torch.Tensor): Price of the derivative.
            precision (float, default=1e-6): Computational precision of the implied volatility.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - max_log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - price: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        s, m, t, p = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, price)
        )
        get_price = lambda volatility: self.price(s, m, t, volatility)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



import abc

import torch


class Feature(abc.ABC):
    """Base class for all features.

    All features should subclass this class.
    """

    @abc.abstractmethod
    def __getitem__(self, i) -> torch.Tensor:
        """Return feature tensor.

        Args:
            i (int): The index of the time step to get the feature.

        Returns:
            torch.Tensor with shape :math:`(N, 1)`, where :math:`N` is the number
            of simulated price paths.
        """

    def of(self, derivative=None, hedger=None):
        """Set `derivative` and `hedger` to the attributes of `self`.

        Args:
            derivative (Derivative, optional): The derivative to compute features.
            hedger (Hedger, optional): The hedger to compute features.

        Returns:
            self
        """
        if not hasattr(self, "derivative") or derivative is not None:
            self.derivative = derivative
        if not hasattr(self, "hedger") or derivative is not None:
            self.hedger = hedger
        return self



import torch
from torch import Tensor


def generate_brownian(
    n_paths: int,
    n_steps: int,
    init_value: float = 0.0,
    volatility: float = 0.2,
    dt: float = 1 / 250,
    dtype: torch.dtype = None,
    device: torch.device = None,
) -> Tensor:
    """Returns time series following the Brownian motion.

    The drift of the time series is assumed to be vanishing.

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_value (float, default=0.0): The initial value of the time series.
        volatility (float, default=0.2): The volatility of the Brownian motion.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If `None`, uses a global default
            (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - Output: :math:`(N, T)`, where :math:`N` is the number of paths and
          :math:`T` is the number of time steps.

    Returns:
        torch.Tensor

    Examples:

        >>> _ = torch.manual_seed(42)
        >>> generate_brownian(2, 5)
        tensor([[ 0.0000,  0.0016,  0.0046,  0.0075, -0.0067],
                [ 0.0000,  0.0279,  0.0199,  0.0257,  0.0291]])
    """
    randn = torch.randn((n_paths, n_steps), dtype=dtype, device=device)
    randn[:, 0] = 0.0
    return init_value + volatility * torch.tensor(dt).to(randn).sqrt() * randn.cumsum(1)


def generate_geometric_brownian(
    n_paths: int,
    n_steps: int,
    init_value: float = 1.0,
    volatility: float = 0.2,
    dt: float = 1 / 250,
    dtype: torch.dtype = None,
    device: torch.device = None,
) -> Tensor:
    """Returns time series following the geometric Brownian motion.

    The drift of the time series is assumed to be vanishing.

    Args:
        n_steps (int): The number of time steps.
        n_paths (int): The number of simulated paths.
        init_value (float, default=0.0): The initial value of the time series.
        volatility (float, default=0.2): The volatility of the Brownian motion.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If `None`, uses a global default
            (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - Output: :math:`(T, N)`, where :math:`T` is the number of time steps and
          :math:`N` is the number of paths.

    Returns:
        torch.Tensor

    Examples:

        >>> _ = torch.manual_seed(42)
        >>> generate_geometric_brownian(2, 5)
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])
    """
    brownian = generate_brownian(
        n_paths=n_paths,
        n_steps=n_steps,
        init_value=0.0,
        volatility=volatility,
        dt=dt,
        dtype=dtype,
        device=device,
    )
    t = dt * torch.arange(n_steps).to(brownian).reshape(1, -1)
    return init_value * torch.exp(brownian - (volatility ** 2) * t / 2)



import torch


class Naked(torch.nn.Module):
    """Returns a tensor filled with the scalar value zero.

    Args:
        out_features (int, default=1): Size of each output sample.

    Shape:
        - Input: :math:`(N, *, H_{\\text{in}})`, where where * means any number of
          additional dimensions and :math:`H_{\\text{in}})` is the number of input
          features.
        - Output: :math:`(N, *, H_{\\text{out}})`, where all but the last dimension
          are the same shape as the input and :math:`H_{\\text{in}})` is
          `out_features`.

    Examples:

        >>> m = Naked()
        >>> input = torch.empty((2, 3))
        >>> m(input)
        tensor([[0.],
                [0.]])
    """

    def __init__(self, out_features=1):
        super().__init__()
        self.out_features = out_features

    def forward(self, input):
        return torch.cat(
            [torch.zeros_like(input[..., :1]) for _ in range(self.out_features)], -1
        )



import torch
from torch import Tensor

from pfhedge._utils.bisect import bisect

from ._base import BSModuleMixin


class BSEuropeanOption(BSModuleMixin):
    """Black-Scholes formula for a European option.

    Args:
        derivative (:class:`pfhedge.instruments.EuropeanOption`, optional):
            The derivative to get the Black-Scholes formula.
        call (bool, default=True): Specifies whether the option is
            call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - Input : :math:`(N, *, 3)`, where :math:`*` means any number
          of additional dimensions. See `inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`. Delta of the derivative.
          All but the last dimension are the same shape as the input.

    .. seealso ::

        - :class:`pfhedge.nn.BlackScholes`:
          Initialize Black-Scholes formula module from a derivative.

    Examples:

        The `forward` method returns delta of the derivative.

        >>> from pfhedge.nn import BSEuropeanOption
        >>> m = BSEuropeanOption()
        >>> m.inputs()
        ['log_moneyness', 'expiry_time', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, 0.1, 0.2],
        ...     [ 0.00, 0.1, 0.2],
        ...     [ 0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[0.4497],
                [0.5126],
                [0.5752]])

        One can instantiate it using a
        :class:`pfhedge.instruments.EuropeanOption`.

        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> deriv = EuropeanOption(BrownianStock(), call=False)
        >>> m = BSEuropeanOption(deriv)
        >>> m
        BSEuropeanOption(call=False)
    """

    def __init__(self, derivative=None, call=True, strike=1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

    def extra_repr(self):
        params = []
        if not self.call:
            params.append(f"call={self.call}")
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def delta(
        self, log_moneyness: Tensor, expiry_time: Tensor, volatility: Tensor
    ) -> Tensor:
        """Returns delta of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        delta = self.N.cdf(self.d1(s, t, v))
        delta = delta - 1 if not self.call else delta  # put-call parity

        return delta

    def gamma(
        self, log_moneyness: Tensor, expiry_time: Tensor, volatility: Tensor
    ) -> Tensor:
        """Returns gamma of the derivative.

        Args:
            log_moneyness: (torch.Tensor): Log moneyness of the underlying asset.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        price = self.strike * torch.exp(s)
        gamma = self.N.pdf(self.d1(s, t, v)) / (price * v * torch.sqrt(t))

        return gamma

    def price(
        self, log_moneyness: Tensor, expiry_time: Tensor, volatility: Tensor
    ) -> Tensor:
        """Returns price of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            max_log_moneyness (torch.Tensor): Cumulative maximum of the log moneyness.
            expiry_time (torch.Tensor): Time to expiry of the option.
            volatility (torch.Tensor): Volatility of the underlying asset.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - volatility: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns:
            Tensor
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        n1 = self.N.cdf(self.d1(s, t, v))
        n2 = self.N.cdf(self.d2(s, t, v))

        price = self.strike * (torch.exp(s) * n1 - n2)

        if not self.call:
            price += self.strike * (torch.exp(s) - 1)  # put-call parity

        return price

    def implied_volatility(
        self, log_moneyness: Tensor, expiry_time: Tensor, price: Tensor, precision=1e-6
    ) -> Tensor:
        """
        Returns implied volatility of the derivative.

        Args:
            log_moneyness (torch.Tensor): Log moneyness of the underlying asset.
            expiry_time (torch.Tensor): Time to expiry of the option.
            price (torch.Tensor): Price of the derivative.
            precision (float): Computational precision of the
                implied volatility.

        Shape:
            - log_moneyness: :math:`(N, *)`
            - expiry_time: :math:`(N, *)`
            - price: :math:`(N, *)`
            - output: :math:`(N, *)`

        Returns
            Tensor
        """
        s, t, p = map(torch.as_tensor, (log_moneyness, expiry_time, price))
        get_price = lambda v: self.price(s, t, v)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



import abc
from inspect import signature

import torch
from torch import Tensor
from torch.distributions.normal import Normal
from torch.nn import Module


class BSModuleMixin(Module):
    """A mixin class for Black-Scholes formula modules.

    Shape:
        - Input: :math:`(N, *, H_\\text{in})`, where :math:`*` means any number of
          additional dimensions. See `inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`: All but the last dimension are the same shape
          as the input.
    """

    def forward(self, input: Tensor) -> Tensor:
        """Returns delta of the derivative.

        Args:
            input (torch.Tensor): The input tensor. Features are concatenated along
                the last dimension.

        Returns:
            torch.Tensor
        """
        return self.delta(*(input[..., [i]] for i in range(input.size()[-1])))

    @abc.abstractmethod
    def delta(*args, **kwargs) -> torch.Tensor:
        """Returns delta of the derivative.

        Returns:
            torch.Tensor
        """

    def inputs(self) -> list:
        """Returns a list of names of input features.

        By default, this method infers the names of features from the signature of
        the `delta` method.
        Please override this method if the signature is different from
        the feature names.
        """
        return list(signature(self.delta).parameters.keys())

    @property
    def N(self) -> Normal:
        """Returns normal distribution with zero mean and unit standard deviation.

        It is almost the same with `torch.distibution.normal.Normal(0, 1)`, but has
        a method `pdf` which stands for the partial distribution function.
        """
        normal = Normal(torch.tensor(0.0), torch.tensor(1.0))
        setattr(normal, "pdf", lambda value: normal.log_prob(value).exp())
        return normal

    @staticmethod
    def d1(log_moneyness: Tensor, expiry_time: Tensor, volatility: Tensor) -> Tensor:
        """Returns :math:`d_1` in the Black-Scholes formula.

        Args:
            log_moneyness (Tensor): Log moneyness of the underlying asset.
            expiry_time (Tensor): Time to expiry of the option.
            volatility (Tensor): Volatility of the underlying asset.

        Returns:
            torch.Tensor
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        return (s + (v ** 2 / 2) * t) / (v * torch.sqrt(t))

    @staticmethod
    def d2(log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """Returns :math:`d_2` in the Black-Scholes formula.

        Args:
            log_moneyness (Tensor): Log moneyness of the underlying asset.
            expiry_time (Tensor): Time to expiry of the option.
            volatility (Tensor): Volatility of the underlying asset.

        Returns:
            torch.Tensor
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        return (s - (v ** 2 / 2) * t) / (v * torch.sqrt(t))



import torch
from torch import Tensor
from torch.nn import Module
from torch.optim import Adam
from tqdm import tqdm

from pfhedge._utils.hook import save_prev_output
from pfhedge._utils.operations import ensemble_mean
from pfhedge.features import get_feature

from .loss import EntropicRiskMeasure
from .loss import HedgeLoss


class Hedger(Module):
    """A `torch.nn.Module` to hedge and price derivatives.

    Args:
        model (torch.nn.Module): Hedging model to compute the hedge ratio at the
            next time step from the input features at the current time step.
            The input and output shapes should be :math:`(N, H_\\text{in})` and
            :math:`(N, 1)` where `N` stands for the number of Monte Carlo paths
            of the asset prices and `H_in` stands for the number of input features
            (namely, `H_in = len(inputs)`).
        inputs (list[str|Feature]): List of (names of) input features to feed to model.
            See `sorted(pfhedge.features.FEATURES)` for valid options.
        criterion (HedgeLoss, default=EntropicRiskMeasure()):
            Loss function to minimize by hedging.
            Default: :class:`pfhedge.nn.EntropicRiskMeasure()` .

    Shape:
        - Input: :math:`(N, H_{\\text{in}})`, where, :math:`H_{\\text{in}}`
          is the number of input features. See `inputs()` for the names of input
          features.
        - Output: :math:`(N, 1)`:

    Examples:


        Black-Scholes' delta hedging strategy.

        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> from pfhedge.nn import BlackScholes
        >>> deriv = EuropeanOption(BrownianStock(cost=1e-4))
        >>> model = BlackScholes(deriv)
        >>> hedger = Hedger(model, model.inputs())
        >>> hedger
        Hedger(
          inputs=['log_moneyness', 'expiry_time', 'volatility'],
          (model): BSEuropeanOption()
          (criterion): EntropicRiskMeasure()
        )
        >>> hedger.compute_pnl(deriv, n_paths=5)
        tensor([..., ..., ..., ..., ...])
        >>> hedger.price(deriv)
        tensor(...)

        Whalley-Wilmott's no-transaction-band strategy.

        >>> from pfhedge.nn import WhalleyWilmott
        >>> model = WhalleyWilmott(deriv)
        >>> hedger = Hedger(model, model.inputs())
        >>> hedger
        Hedger(
          inputs=['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge'],
          (model): WhalleyWilmott(
            (bs): BSEuropeanOption()
            (clamp): Clamp()
          )
          (criterion): EntropicRiskMeasure()
        )
        >>> hedger.compute_pnl(deriv, n_paths=5)
        tensor([..., ..., ..., ..., ...])
        >>> hedger.price(deriv)
        tensor(...)

        A naked position (never hedge at all).

        >>> from pfhedge.nn import Naked
        >>> hedger = Hedger(Naked(), ["zero"])
        >>> hedger.compute_pnl(deriv, n_paths=5)
        tensor([..., ..., ..., ..., ...])
        >>> hedger.price(deriv)
        tensor(...)

        A strategy represented by a neural network (Deep Hedging).

        >>> from pfhedge.nn import MultiLayerPerceptron
        >>> model = MultiLayerPerceptron()
        >>> hedger = Hedger(model, ["moneyness", "expiry_time", "volatility"])
        >>> _ = hedger.compute_pnl(deriv, n_paths=1)  # lazily derermine in_features
        >>> hedger
        Hedger(
          inputs=['moneyness', 'expiry_time', 'volatility'],
          (model): MultiLayerPerceptron(
            (0): Linear(in_features=3, out_features=32, bias=True)
            (1): ReLU()
            (2): Linear(in_features=32, out_features=32, bias=True)
            (3): ReLU()
            (4): Linear(in_features=32, out_features=32, bias=True)
            (5): ReLU()
            (6): Linear(in_features=32, out_features=32, bias=True)
            (7): ReLU()
            (8): Linear(in_features=32, out_features=1, bias=True)
            (9): Identity()
          )
          (criterion): EntropicRiskMeasure()
        )
        >>> history = hedger.fit(deriv, verbose=False, n_paths=1, n_epochs=1)
        >>> hedger.compute_pnl(deriv, n_paths=3)
        tensor([..., ..., ...], grad_fn=<SubBackward0>)
        >>> hedger.price(deriv)
        tensor(...)
    """

    def __init__(
        self, model: Module, inputs: list, criterion: HedgeLoss = EntropicRiskMeasure()
    ):
        super().__init__()

        self.model = model
        self.inputs = [get_feature(i) for i in inputs]
        self.criterion = criterion

        # This hook saves the hedger's previous output to an attribute `prev`.
        self.register_forward_hook(save_prev_output)

    def forward(self, input: Tensor) -> Tensor:
        """Returns the outout of `model`.

        The output is supposed to represent the hedge ratio at the next time step.
        """
        return self.model(input)

    def extra_repr(self) -> str:
        params = []
        if not isinstance(self.model, torch.nn.Module):
            params.append(f"model={self.model.__name__},")
        params.append(f"inputs={[str(f) for f in self.inputs]},")

        return "\n".join(params)

    def compute_pnl(
        self, derivative, n_paths: int = 1000, init_price: float = 1.0
    ) -> Tensor:
        """Returns the profit and loss distribution after hedging.

        A hedger sells the derivative to its customer and obliges to settle the payoff
        at maturity. The dealer hedges the risk of this liability by trading
        the underlying instrument of the derivative based on `model`.
        The resulting profit and loss is obtained by adding up the payoff to the
        customer, capital gains from the underlying asset, and the transaction cost.

        Args:
            derivative (pfhedge.instruments.Derivative): The derivative to hedge.
            n_paths (int, default=1000): The number of simulated price paths of the
                underlying instrument.
            init_price (float, default=1.0): The initial price of the underlying
                instrument of the derivative.

        Shape:
            - Output: :math:`(N)`, where :math:`N` is the number of paths.

        Returns:
            torch.Tensor
        """
        self.inputs = [feature.of(derivative, self) for feature in self.inputs]

        derivative.simulate(n_paths=n_paths, init_price=init_price)
        # cashflow: shape (N, T - 1)
        cashflow = (
            derivative.underlier.prices[..., 1:] - derivative.underlier.prices[..., :-1]
        )

        # prev_output: shape (N)
        self.register_buffer(
            "prev_output",
            torch.zeros_like(derivative.underlier.prices[..., :1]),
            persistent=False,
        )
        pnl = 0

        # Simulate hedging over time.
        n_steps = derivative.underlier.prices.size(1)  # = T
        for i in range(n_steps - 1):
            prev_hedge = self.get_buffer("prev_output").reshape(-1)

            # Compute the hedge ratio at the next time step.
            hedge = self(torch.cat([f[i] for f in self.inputs], 1)).reshape(-1)

            # Receive profit and loss from the underlying asset.
            pnl += hedge * cashflow[..., i]
            # Deduct transactoon cost.
            pnl -= (
                derivative.underlier.cost
                * torch.abs(hedge - prev_hedge)
                * derivative.underlier.prices[..., i]
            )

        # Settle the derivative's payoff.
        pnl -= derivative.payoff()

        # Delete the attribute `prev_output` in case a hedger has a feature
        # `PrevHedge` and one calls `compute_pnl` twice.
        # If `prev_output` is not deleted, `prev_output` at the last time step
        # in the first call could be referred to by `PrevHedge` at the first
        # time step at the second call,
        # which results in an unexpected output (while we expect zeros).
        if hasattr(self, "prev_output"):
            delattr(self, "prev_output")

        return pnl

    def compute_loss(
        self,
        derivative,
        n_paths: int = 1000,
        n_times: int = 1,
        init_price: float = 1.0,
        enable_grad: bool = True,
    ) -> Tensor:
        """Returns the loss of the profit and loss distribution after hedging.

        Args:
            derivative (pfhedge.instruments.Derivative): The derivative to hedge.
            n_paths (int, default=1000): The number of simulated price paths of the
                underlying instrument.
            n_times (int, default=1): If `n_times > 1`, returns the ensemble mean
                of the losses computed through multiple simulations.
            init_price (float, default=1.0): The initial price of the underlying
                instrument of the derivative.
            enable_grad (bool, default=True): Context-manager that sets gradient
                calculation to on or off.

        Shape:
            - Output: :math:`()`

        Returns:
            torch.Tensor
        """
        with torch.set_grad_enabled(enable_grad):
            loss = lambda: self.criterion(
                self.compute_pnl(derivative, init_price=init_price, n_paths=n_paths)
            )
            mean_loss = ensemble_mean(loss, n_times=n_times)

        return mean_loss

    def fit(
        self,
        derivative,
        n_epochs: int = 100,
        n_paths: int = 1000,
        n_times: int = 1,
        optimizer=Adam,
        init_price: float = 1.0,
        verbose: bool = True,
    ) -> list:
        """Train the hedging model to hedge the given derivative.

        It returns the trade history, that is, validation loss after each simulation.

        Args:
            derivative (pfhedge.instruments.Derivative): The derivative to hedge.
            n_epochs (int, default=100): Number of Monte-Carlo simulations.
            n_paths (int, default=1000): The number of simulated price paths of the
                underlying instrument.
            n_times (int, default=1): If `n_times > 1`, returns the ensemble mean of
                the losses computed through multiple simulations.
            optimizer (torch.optim.Optimizer, default=Adam): The optimizer algorithm
                to use.  It can be an instance or a class of `torch.optim.Optimizer`.
            init_price (float, default=1.0): The initial price of the underlying
                instrument of the derivative.
            verbose (bool, default=True): If `True`, print progress of the training to
                standard output.

        Returns:
            list[float]

        Examples:

            Here is an example to use a custom optimizer.

            >>> from pfhedge.instruments import BrownianStock
            >>> from pfhedge.instruments import EuropeanOption
            >>> from pfhedge.nn import MultiLayerPerceptron
            >>> from torch.optim import SGD
            >>> derivative = EuropeanOption(BrownianStock())
            >>> hedger = Hedger(MultiLayerPerceptron(), ["zero"])
            >>> # Run a "dummy" forward to initialize lazy parameters
            >>> _ = hedger.compute_pnl(derivative, n_paths=1)
            >>> _ = hedger.fit(
            ...     derivative,
            ...     optimizer=SGD(hedger.model.parameters(), lr=0.1),
            ...     n_epochs=1,
            ...     verbose=False)

            You can also pass a class object of an optimizer.

            >>> from torch.optim import Adadelta
            >>> derivative = EuropeanOption(BrownianStock())
            >>> hedger = Hedger(MultiLayerPerceptron(), ["zero"])
            >>> _ = hedger.fit(
            ...     derivative,
            ...     optimizer=Adadelta,
            ...     n_epochs=1,
            ...     verbose=False)
        """
        if isinstance(optimizer, type):
            # Run a "dummy" forward to initialize lazy parameters
            _ = self.compute_pnl(derivative, n_paths=1)
            optimizer = optimizer(self.model.parameters())
            if not isinstance(optimizer, torch.optim.Optimizer):
                raise TypeError("optimizer is not torch.optim.Optimizer")

        compute_loss = lambda **kwargs: self.compute_loss(
            derivative, n_paths=n_paths, init_price=init_price, **kwargs
        )

        history = []
        progress = tqdm(range(n_epochs)) if verbose else range(n_epochs)
        for _ in progress:
            # Compute training loss and backpropagate
            self.train()
            optimizer.zero_grad()
            loss = compute_loss()
            loss.backward()
            optimizer.step()

            # Compute validation loss
            self.eval()
            loss = compute_loss(n_times=n_times, enable_grad=False)
            history.append(loss.item())

            if verbose:
                progress.desc = f"Loss={loss:.5e}"

        return history

    def price(
        self,
        derivative,
        n_paths: int = 1000,
        n_times: int = 1,
        init_price: float = 1.0,
        enable_grad: bool = False,
    ) -> Tensor:
        """
        Evaluate the premium of the given derivative.

        Args:
            derivative (pfhedge.instuments.Derivative): The derivative to price.
            n_paths (int, default=1000): The number of simulated price paths of the
                underlying instrument.
            n_times (int, default=1): If `n_times > 1`, returns the ensemble mean of
                the losses computed through multiple simulations.
            init_price (float, default=1.0): The initial price of the underlying
                instrument of the derivative.
            enable_grad (bool, default=False): Context-manager that sets gradient
                calculation to on or off.

        Shape:
            - Output: :math:`()`

        Returns:
            torch.Tensor
        """
        with torch.set_grad_enabled(enable_grad):
            # Negative because selling
            price = lambda: -self.criterion.cash(
                self.compute_pnl(derivative, init_price=init_price, n_paths=n_paths)
            )
            mean_price = ensemble_mean(price, n_times=n_times)

        return mean_price



import torch
from torch import Tensor

from ..functional import clamp
from ..functional import leaky_clamp


class LeakyClamp(torch.nn.Module):
    """Leakily Clamp all elements in `input` into the range :math:`[\\min, \\max]`.

    The bounds :math:`\\min` and :math:`\\max` can be tensors.

    If :math:`\\min \\leq \\max`:

    .. math ::

        \\text{output} = \\begin{cases}
        \\min + \\text{clampled_slope} * (\\text{input} - \\min) &
        \\text{input} < \\min \\\\
        \\text{input} & \\min \\leq \\text{input} \\leq \\max \\\\
        \\max + \\text{clampled_slope} * (\\text{input} - \\max) &
        \\max < \\text{input}
        \\end{cases}

    Otherwise:

    .. math ::

        \\text{output} = \\frac12 (\\min + \\max)

    Args:
        clamped_slope (float, default=0.01):
            Controls the slope in the clampled regions.

    Shape:
        - input: :math:`(N, *)`, where :math:`*` means any number of additional
          dimensions.
        - min: :math:`(N, *)`
        - max: :math:`(N, *)`
        - output: :math:`(N, *)`, same shape as the input.

    Examples:

        >>> from pfhedge.nn import LeakyClamp
        >>> m = LeakyClamp()
        >>> input = torch.linspace(-2, 12, 15) * 0.1
        >>> input
        tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
                 0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
        >>> min = torch.tensor(0.0)
        >>> max = torch.tensor(1.0)
        >>> m(input, min, max)
        tensor([-2.0000e-03, -1.0000e-03,  0.0000e+00,  1.0000e-01,  2.0000e-01,
                 3.0000e-01,  4.0000e-01,  5.0000e-01,  6.0000e-01,  7.0000e-01,
                 8.0000e-01,  9.0000e-01,  1.0000e+00,  1.0010e+00,  1.0020e+00])
    """

    def __init__(self, clamped_slope: float = 0.01):
        super().__init__()
        self.clamped_slope = clamped_slope

    def extra_repr(self):
        return f"clamped_slope={self.clamped_slope}" if self.clamped_slope != 0 else ""

    def forward(
        self, input: Tensor, min_value: Tensor = None, max_value: Tensor = None
    ) -> Tensor:
        """Clamp all elements in `input` into the range :math:`[\\min, \\max]`.

        Args:
            input (torch.Tensor): The input tensor.
            min (torch.Tensor, optional): Lower-bound of the range to be clamped to.
            max (torch.Tensor, optional): Upper-bound of the range to be clamped to.

        Shape:
            - input: :math:`(N, *)`, where :math:`*` means any number of additional
              dimensions.
            - min: :math:`(N, *)`
            - max: :math:`(N, *)`
            - output: :math:`(N, *)`, same shape as the input.

        Returns:
            torch.Tensor
        """
        return leaky_clamp(
            input, min_value, max_value, clamped_slope=self.clamped_slope
        )


class Clamp(torch.nn.Module):
    """Clamp all elements in `input` into the range :math:`[\\min, \\max]`.

    The bounds :math:`\\min` and :math:`\\max` can be tensors.

    If :math:`\\min \\leq \\max`:

    .. math ::

        \\text{output} = \\begin{cases}
        \\min & \\text{input} < \\min \\\\
        \\text{input} & \\min \\leq \\text{input} \\leq \\max \\\\
        \\max & \\max < \\text{input}
        \\end{cases}

    Otherwise:

    .. math ::

        \\text{output} = \\frac12 (\\min + \\max)

    Shape:
        - input: :math:`(N, *)`, where :math:`*` means any number of additional
          dimensions.
        - min: :math:`(N, *)`
        - max: :math:`(N, *)`
        - output: :math:`(N, *)`, same shape as the input.

    Examples:

        >>> from pfhedge.nn import Clamp
        >>> m = Clamp()
        >>> input = torch.linspace(-2, 12, 15) * 0.1
        >>> input
        tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
                 0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
        >>> min = torch.tensor(0.0)
        >>> max = torch.tensor(1.0)
        >>> m(input, 0.0, 1.0)
        tensor([0.0000, 0.0000, 0.0000, 0.1000, 0.2000, 0.3000, 0.4000, 0.5000, 0.6000,
                0.7000, 0.8000, 0.9000, 1.0000, 1.0000, 1.0000])

        When :math:`\\min > \\max`, returns the mean of :math:`\\min` and :math:`\\max`.

        >>> input = torch.tensor([1.0, 0.0])
        >>> min = torch.tensor([0.0, 1.0])
        >>> max = torch.tensor([0.0, 0.0])
        >>> m(input, min, max)
        tensor([0.0000, 0.5000])
    """

    def forward(self, input: Tensor, min: Tensor = None, max: Tensor = None) -> Tensor:
        """Clamp all elements in `input` into the range :math:`[\\min, \\max]`.

        Args:
            input (torch.Tensor): The input tensor.
            min (torch.Tensor, optional): Lower-bound of the range to be clamped to.
            max (torch.Tensor, optional): Upper-bound of the range to be clamped to.

        Shape:
            - input: :math:`(N, *)`, where :math:`*` means any number of additional
              dimensions.
            - min: :math:`(N, *)`
            - max: :math:`(N, *)`
            - output: :math:`(N, *)`, same shape as the input.

        Returns:
            torch.Tensor
        """
        return clamp(input, min, max)



from ..stochastic import generate_geometric_brownian
from ._base import Primary


class BrownianStock(Primary):
    """A stock of which prices follow the geometric Brownian motion.

    The drift of the prices is assumed to be vanishing.

    Args:
        volatility (float, default=0.2): The volatility of the price.
        cost (float, default=0.0): The transaction cost rate.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.device, optional): Desired device of returned tensor.
            Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Attributes:
        prices (torch.Tensor): The prices of the instrument.
            This attribute is supposed to set by a method `simulate()`.
            The shape is :math:`(T, N)`. Here, :math:`T` is the number of time steps and
            :math:`N` is the number of simulated paths.

    Examples:

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> _ = torch.manual_seed(42)
        >>> stock = BrownianStock(volatility=0.20)
        >>> stock.simulate(time_horizon=5 / 250, n_paths=2)
        >>> stock.prices
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])

        Using custom `dtype` and `device`.

        >>> stock = BrownianStock()
        >>> stock.to(dtype=torch.float64, device="cuda:0")
        BrownianStock(..., dtype=torch.float64, device='cuda:0')
    """

    def __init__(
        self,
        volatility: float = 0.2,
        cost: float = 0.0,
        dt: float = 1 / 250,
        dtype=None,
        device=None,
    ):
        super().__init__()

        self.volatility = volatility
        self.cost = cost
        self.dt = dt

        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"volatility={self.volatility:.2e}"]
        if self.cost != 0.0:
            params.append(f"cost={self.cost:.2e}")
        params.append(f"dt={self.dt:.2e}")
        params += self.dinfo
        return self.__class__.__name__ + "(" + ", ".join(params) + ")"

    def simulate(
        self, time_horizon: float, n_paths: int = 1, init_price: float = 1.0
    ) -> None:
        """
        Simulate time series of prices and set an attribute `prices`.

        Args:
            time_horizon (float): The period of time to simulate the price.
            n_paths (int, default=1): The number of paths to simulate.
            init_price (float, default=1.0): The initial value of the prices.

        Returns:
            None
        """
        n_steps = int(time_horizon / self.dt)
        self.prices = generate_geometric_brownian(
            n_paths=n_paths,
            n_steps=n_steps,
            init_value=init_price,
            volatility=self.volatility,
            dt=self.dt,
            dtype=self.dtype,
            device=self.device,
        )



from copy import deepcopy

import torch
from torch.nn import Identity
from torch.nn import LazyLinear
from torch.nn import Module
from torch.nn import ReLU
from torch.nn import Sequential


class MultiLayerPerceptron(Sequential):
    """A feed-forward neural network.

    Number of input features is lazily determined.

    Args:
        out_features (int, default=1): Size of each output sample.
        n_layers (int, default=4): Number of hidden layers.
        n_units (int or tuple[int], default=32): Number of units in each hidden layer.
            If `tuple[int]`, it specifies different number of units for each layer.
        activation (torch.nn.Module, default=torch.nn.ReLU()):
            Activation module of the hidden layers.
        out_activation (torch.nn.Module, default=torch.nn.Identity()):
            Activation module of the output layer.

    Shape:
        - Input: :math:`(N, *, H_{\\text{in}})`, where where * means any number of
          additional dimensions and :math:`H_{\\text{in}})` is the number of input
          features.
        - Output: :math:`(N, *, H_{\\text{out}})`, where all but the last dimension
          are the same shape as the input and :math:`H_{\\text{in}})` is
          `out_features`.

    Examples:

        >>> from pfhedge.nn import MultiLayerPerceptron
        >>> m = MultiLayerPerceptron()
        >>> _ = m(torch.empty((1, 2)))  # lazily determine the number of input features
        >>> m
        MultiLayerPerceptron(
          (0): Linear(in_features=2, out_features=32, bias=True)
          (1): ReLU()
          (2): Linear(in_features=32, out_features=32, bias=True)
          (3): ReLU()
          (4): Linear(in_features=32, out_features=32, bias=True)
          (5): ReLU()
          (6): Linear(in_features=32, out_features=32, bias=True)
          (7): ReLU()
          (8): Linear(in_features=32, out_features=1, bias=True)
          (9): Identity()
        )
        >>> input = torch.randn((3, 2))
        >>> m(input)
        tensor([[...],
                [...],
                [...]], grad_fn=<AddmmBackward>)

        Specify different number of layers for each layer:

        >>> m = MultiLayerPerceptron(n_layers=2, n_units=(16, 32))
        >>> _ = m(torch.empty(1, 1))
        >>> m
        MultiLayerPerceptron(
          (0): Linear(in_features=1, out_features=16, bias=True)
          (1): ReLU()
          (2): Linear(in_features=16, out_features=32, bias=True)
          (3): ReLU()
          (4): Linear(in_features=32, out_features=1, bias=True)
          (5): Identity()
        )
    """

    def __init__(
        self,
        out_features: int = 1,
        n_layers: int = 4,
        n_units=32,
        activation: Module = ReLU(),
        out_activation: Module = Identity(),
    ):
        n_units = (n_units,) * n_layers if isinstance(n_units, int) else n_units

        layers = []
        for i in range(n_layers):
            layers.append(LazyLinear(n_units[i]))
            layers.append(deepcopy(activation))
        layers.append(LazyLinear(out_features))
        layers.append(deepcopy(out_activation))

        super().__init__(*layers)



import torch

from .american_binary import BSAmericanBinaryOption
from .european import BSEuropeanOption
from .european_binary import BSEuropeanBinaryOption
from .lookback import BSLookbackOption


class BlackScholes(torch.nn.Module):
    """Initialize Black-Scholes formula module from a derivative.

    The `forward` method returns the Black-Scholes delta.

    Args:
        derivative (:class:`pfhedge.instruments.Derivative`):
            The derivative to get the Black-Scholes formula.

    Shape:
        - Input : :math:`(N, *, H_{\\mathrm{in}})`, where :math:`*` means any number of
          additional dimensions and :math:`H_{\\mathrm{in}}` is the number of input
          features. See `inputs()` for the names of input features.
        - Output : :math:`(N, *, 1)`. All but the last dimension are the same shape
          as the input.

    Examples:

        One can instantiate Black-Scholes module by using a derivative.
        For example, one can instantiate :class:`BSEuropeanOption` using
        a :class:`pfhedge.instruments.EuropeanOption`.
        The `forward` method returns delta of the derivative.

        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> from pfhedge.nn import BlackScholes
        >>> deriv = EuropeanOption(BrownianStock())
        >>> m = BlackScholes(deriv)
        >>> m
        BSEuropeanOption()
        >>> m.inputs()
        ['log_moneyness', 'expiry_time', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, 0.1, 0.2],
        ...     [ 0.00, 0.1, 0.2],
        ...     [ 0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[0.4497],
                [0.5126],
                [0.5752]])

        Instantiating :class:`BSLookbackOption` using a
        :class:`pfhedge.instruments.LookbackOption`.

        >>> from pfhedge.instruments import LookbackOption
        >>> deriv = LookbackOption(BrownianStock(), strike=1.03)
        >>> m = BlackScholes(deriv)
        >>> m
        BSLookbackOption(strike=1.03)
        >>> m.inputs()
        ['log_moneyness', 'max_log_moneyness', 'expiry_time', 'volatility']
        >>> input = torch.tensor([
        ...     [-0.01, -0.01, 0.1, 0.2],
        ...     [ 0.00,  0.00, 0.1, 0.2],
        ...     [ 0.01,  0.01, 0.1, 0.2]])
        >>> m(input)
        tensor([[0.9208],
                [1.0515],
                [1.0515]])
    """

    inputs: list

    def __init__(self, derivative):
        self.__class__ = {
            "EuropeanOption": BSEuropeanOption,
            "LookbackOption": BSLookbackOption,
            "AmericanBinaryOption": BSAmericanBinaryOption,
            "EuropeanBinaryOption": BSEuropeanBinaryOption,
        }[derivative.__class__.__name__]
        self.__init__(derivative)



import torch
from torch import Tensor


def bisect(
    function, target: Tensor, lower, upper, precision: float = 1e-6, abort: int = 100000
) -> Tensor:
    """Perform binary search over a tensor.

    The output tensor approximately satisfies the following relation:

    .. code-block ::

        function(output) = target

    Args:
        function (callable[[Tensor], Tensor]): Monotone increasing or decreasing
            function.
        target (Tensor): Target of function values.
        lower (Tensor or float): Lower bound of binary search.
        upper (Tensor or float): Upper bound of binary search.
        precision (float, default=1e-6): Precision of output.
        abort (int, default 100000): If the number of iterations exceeds this
            value, abort computation and raise RuntimeError.

    Returns:
        torch.Tensor

    Raises:
        RuntimeError: If the number of iteration exceeds `abort`.

    Examples:

        >>> target = torch.tensor([-1.0, 0.0, 1.0])
        >>> function = torch.log
        >>> output = bisect(function, target, 0.01, 10.0)
        >>> output
        tensor([0.3679, 1.0000, 2.7183])
        >>> torch.allclose(function(output), target, atol=1e-6)
        True

        Monotone decreasing function:

        >>> function = lambda x: -torch.log(x)
        >>> output = bisect(function, target, 0.01, 10.0)
        >>> output
        tensor([2.7183, 1.0000, 0.3679])
        >>> torch.allclose(function(output), target, atol=1e-6)
        True
    """
    lower, upper = map(torch.as_tensor, (lower, upper))

    if not (lower < upper).all():
        raise ValueError("condition lower < upper should be satisfied.")

    if (function(lower) > function(upper)).all():
        # If decreasing function
        mf = lambda x: -function(x)
        return bisect(mf, -target, lower, upper, precision=precision, abort=abort)

    lower = torch.full_like(target, lower)
    upper = torch.full_like(target, upper)

    n_iter = 0
    while torch.max(upper - lower) > precision:
        n_iter += 1
        if n_iter > abort:
            raise RuntimeError(f"Aborting since iteration exceeds abort={abort}.")

        m = (lower + upper) / 2
        output = function(m)
        lower = torch.where(output < target, m, lower)
        upper = torch.where(output >= target, m, upper)

    return upper



from math import ceil

import torch
import torch.nn.functional as fn
from torch import Tensor


def european_payoff(input: Tensor, call: bool = True, strike: float = 1.0) -> Tensor:
    """Returns the payoff of a European option.

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)`, where, :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return fn.relu(input[..., -1] - strike)
    else:
        return fn.relu(strike - input[..., -1])


def lookback_payoff(input: Tensor, call: bool = True, strike: float = 1.0) -> Tensor:
    """Returns the payoff of a lookback option with a fixed strike.

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)`, where, :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return fn.relu(input.max(dim=-1).values - strike)
    else:
        return fn.relu(strike - input.min(dim=-1).values)


def american_binary_payoff(
    input: Tensor, call: bool = True, strike: float = 1.0
) -> Tensor:
    """
    Returns the payoff of an American binary option.

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)`, where, :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return (input.max(dim=-1).values >= strike).to(input)
    else:
        return (input.min(dim=-1).values <= strike).to(input)


def european_binary_payoff(
    input: Tensor, call: bool = True, strike: float = 1.0
) -> Tensor:
    """
    Returns the payoff of a European binary option.

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)`, where, :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return (input[..., -1] >= strike).to(input)
    else:
        return (input[..., -1] <= strike).to(input)


def exp_utility(input: Tensor, a: float = 1.0) -> Tensor:
    """Applies an exponential utility function.

    An exponential utility function is defined as:

    .. math ::

        u(x) = -\\exp(-a x) \,.

    Args:
        input (torch.Tensor): The input tensor.
        a (float, default=1.0): The risk aversion coefficient of the exponential
            utility.

    Returns:
        torch.Tensor
    """
    return -torch.exp(-a * input)


def isoelastic_utility(input: Tensor, a: float = 0.5) -> Tensor:
    """Applies an isoelastic utility function.

    An isoelastic utility function is defined as:

    .. math ::

        u(x) = \\begin{cases}
        x^{1 - a} & a \\neq 1 \\\\
        \\log{x} & a = 1
        \\end{cases} \,.

    Args:
        input (torch.Tensor): The input tensor.
        a (float, default=0.5): Relative risk aversion coefficient of the isoelastic
            utility.

    Returns:
        torch.Tensor
    """
    if a == 1.0:
        return torch.log(input)
    else:
        return torch.pow(input, exponent=1.0 - a)


def topp(input, p: float, dim: int = None, largest: bool = True):
    """Returns the largest `p * N` elements of the given input tensor,
    where `N` stands for the total number of elements in the input tensor.

    If `dim` is not given, the last dimension of the `input` is chosen.

    If `largest` is `False` then the smallest elements are returned.

    A namedtuple of `(values, indices)` is returned, where the `indices` are the indices
    of the elements in the original `input` tensor.

    Args:
        input (torch.Tensor): The input tensor.
        p (float): Quantile level.
        dim (int, optional): The dimension to sort along.
        largest (bool, default=True): Controls whether to return largest or smallest
            elements.

    Returns:
        torch.Tensor

    Examples:

        >>> input = torch.arange(1.0, 6.0)
        >>> input
        tensor([1., 2., 3., 4., 5.])
        >>> topp(input, 3 / 5)
        torch.return_types.topk(
        values=tensor([5., 4., 3.]),
        indices=tensor([4, 3, 2]))
    """
    if dim is None:
        return torch.topk(input, ceil(p * input.numel()), largest=largest)
    else:
        return torch.topk(input, ceil(p * input.size()[dim]), dim=dim, largest=largest)


def expected_shortfall(input: torch.Tensor, p: float, dim=None) -> torch.Tensor:
    """Returns the expected shortfall of the given input tensor.

    Args:
        input (torch.Tensor): The input tensor.
        p (float): Quantile level.
        dim (int, optional): The dimension to sort along.
        largest (bool, default=True): Controls whether to return largest or smallest
            elements.

    Examples:

        >>> input = -torch.arange(1., 6.)
        >>> expected_shortfall(input, 3 / 5)
        tensor(4.)

    Returns:
        torch.Tensor
    """
    if dim is None:
        return -topp(input, p=p, largest=False).values.mean()
    else:
        return -topp(input, p=p, largest=False, dim=dim).values.mean(dim=dim)


def leaky_clamp(
    input: Tensor, min: Tensor = None, max: Tensor = None, clamped_slope: float = 0.01
) -> Tensor:
    """Leakily clamp all elements in `input` into the range :math:`[\\min, \\max]`.

    The bounds :math:`\\min` and :math:`\\max` can be tensors.

    See :class:`pfhedge.nn.LeakyClamp` for details.
    """
    x = input

    if min is not None:
        min = torch.as_tensor(min)
        x = torch.max(x, min + clamped_slope * (x - min))

    if max is not None:
        max = torch.as_tensor(max)
        x = torch.min(x, max + clamped_slope * (x - max))

    if min is not None and max is not None:
        x = torch.where(min <= max, x, (min + max) / 2)

    return x


def clamp(input: Tensor, min: Tensor = None, max: Tensor = None) -> Tensor:
    """Clamp all elements in `input` into the range :math:`[\\min, \\max]`.

    The bounds :math:`\\min` and :math:`\\max` can be tensors.

    See :class:`pfhedge.nn.Clamp` for details.
    """
    return leaky_clamp(input, min=min, max=max, clamped_slope=0.0)



from torch import Tensor

from ..nn.functional import lookback_payoff
from ._base import Derivative


class LookbackOption(Derivative):
    """A lookback option with fixed strike.

    A lookback call option provides its holder the right to buy an underlying with
    the strike price and to sell with the highest price until the date of maturity.

    A lookback put option provides its holder the right to sell an underlying with
    the strike price and to buy with the lowest price until the date of maturity.

    The payoff of a lookback call option is given by:

    .. math::

        \\mathrm{payoff} = \\max(\\mathrm{Max} - K, 0)

    Here, :math:`\\mathrm{Max}` is the maximum of the underlying asset's price
    until maturity and :math:`K` is the strike price (`strike`) of the option.

    The payoff of a lookback put option is given by:

    .. math::

        \\mathrm{payoff} = \\max(K - \\mathrm{Min}, 0)

    Here, :math:`\\mathrm{Min}` is the minimum of the underlying asset's price.

    Args:
        underlier (:class:`Primary`): The underlying instrument of the option.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.
        maturity (float, default=20/250): The maturity of the option.
        dtype (torch.device, optional): Desired device of returned tensor.
            Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Examples:

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import LookbackOption
        >>> _ = torch.manual_seed(42)
        >>> deriv = LookbackOption(BrownianStock(), maturity=5/250)
        >>> deriv.simulate(n_paths=2)
        >>> deriv.underlier.prices
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])
        >>> deriv.payoff()
        tensor([0.0073, 0.0292])
    """

    def __init__(
        self,
        underlier,
        call: bool = True,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        dtype=None,
        device=None,
    ):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")
        params += self.dinfo
        return self.__class__.__name__ + "(" + ", ".join(params) + ")"

    def payoff(self) -> Tensor:
        return lookback_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



from torch import Tensor

from ..nn.functional import european_payoff
from ._base import Derivative


class EuropeanOption(Derivative):
    """A European option.

    A European option provides its holder the right to buy (for call option)
    or sell (for put option) an underlying asset with the strike price
    on the date of maturity.

    The payoff of a European call option is given by:

    .. math::

        \\mathrm{payoff} = \\max(S - K, 0)

    Here, :math:`S` is the underlying asset's price at maturity and
    :math:`K` is the strike price (`strike`) of the option.

    The payoff of a European put option is given by:

    .. math::

        \\mathrm{payoff} = \\max(K - S, 0)

    Args:
        underlier (:class:`Primary`): The underlying instrument of the option.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.
        maturity (float, default=20/250): The maturity of the option.
        dtype (torch.device, optional): Desired device of returned tensor.
            Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Examples:

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> _ = torch.manual_seed(42)
        >>> deriv = EuropeanOption(BrownianStock(), maturity=5/250)
        >>> deriv.simulate(n_paths=2)
        >>> deriv.underlier.prices
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])
        >>> deriv.payoff()
        tensor([0.0000, 0.0292])

        Using custom `dtype` and `device`.

        >>> deriv = EuropeanOption(BrownianStock())
        >>> deriv.to(dtype=torch.float64, device="cuda:0")
        EuropeanOption(BrownianStock(...), ..., dtype=torch.float64, device='cuda:0')
    """

    def __init__(
        self,
        underlier,
        call: bool = True,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        dtype=None,
        device=None,
    ):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")
        params += self.dinfo
        return self.__class__.__name__ + "(" + ", ".join(params) + ")"

    def payoff(self) -> Tensor:
        return european_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



import torch

from .bs.bs import BlackScholes
from .clamp import Clamp


class WhalleyWilmott(torch.nn.Module):
    """Initialize Whalley-Wilmott's hedging strategy of a derivative.

    The `forward` method returns the next hedge ratio.

    This is the optimal hedging strategy for asymptotically small transaction cost.

    Args:
        derivative (:class:`pfhedge.instruments.Derivative`): Derivative to hedge.
        a (float, default=1.0): Risk aversion parameter in exponential utility.

    Shape:
        - Input: :math:`(N, *, H_{\\text{in}})`.  Here, :math:`*` means any number of
          additional dimensions and `H_in` is the number of input features.
          See `inputs()` for the names of input features.
        - Output: :math:`(N, *, 1)`. The hedge ratio at the next time step.

    Examples:

        An example for :class:`pfhedge.instruments.EuropeanOption`.

        >>> from pfhedge.nn import WhalleyWilmott
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> deriv = EuropeanOption(BrownianStock(cost=1e-5))
        >>> m = WhalleyWilmott(deriv)
        >>> m.inputs()
        ['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge']
        >>> input = torch.tensor([
        ...     [-0.05, 0.1, 0.2, 0.5],
        ...     [-0.01, 0.1, 0.2, 0.5],
        ...     [ 0.00, 0.1, 0.2, 0.5],
        ...     [ 0.01, 0.1, 0.2, 0.5],
        ...     [ 0.05, 0.1, 0.2, 0.5]])
        >>> m(input)
        tensor([[0.2946],
                [0.5000],
                [0.5000],
                [0.5000],
                [0.7284]])

        An example for :class:`pfhedge.instruments.EuropeanOption` without cost.

        >>> deriv = EuropeanOption(BrownianStock())
        >>> m = WhalleyWilmott(deriv)
        >>> m.inputs()
        ['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge']
        >>> input = torch.tensor([
        ...     [-0.05, 0.1, 0.2, 0.5],
        ...     [-0.01, 0.1, 0.2, 0.5],
        ...     [ 0.00, 0.1, 0.2, 0.5],
        ...     [ 0.01, 0.1, 0.2, 0.5],
        ...     [ 0.05, 0.1, 0.2, 0.5]])
        >>> m(input)
        tensor([[0.2239],
                [0.4497],
                [0.5126],
                [0.5752],
                [0.7945]])

    References:
        - Whalley, A.E. and Wilmott, P., An asymptotic analysis of an optimal hedging model
          for option pricing with transaction costs. Mathematical Finance, 1997, 7, 307–324.
    """

    def __init__(self, derivative, a: float = 1.0):
        super().__init__()
        self.derivative = derivative
        self.a = a

        self.bs = BlackScholes(derivative)
        self.clamp = Clamp()

    def inputs(self):
        return self.bs.inputs() + ["prev_hedge"]

    def extra_repr(self):
        return f"a={self.a}" if self.a != 1 else ""

    def forward(self, input) -> torch.Tensor:
        prev_hedge = input[..., [-1]]

        delta = self.bs(input[..., :-1])
        width = self.width(input[..., :-1])

        return self.clamp(prev_hedge, min=delta - width, max=delta + width)

    def width(self, input) -> torch.Tensor:
        """Returns half-width of the no-transaction band.

        Args:
            input : Tensor

        Shape:
            - Input: :math:`(N, *, H_{\\text{in}} - 1)`
            - Output: :math:`(N, *, 1)`

        Returns:
            torch.Tensor
        """
        cost = self.derivative.underlier.cost

        price = self.derivative.strike * torch.exp(input[..., [0]])
        gamma = self.bs.gamma(*(input[..., [i]] for i in range(input.size()[-1])))
        width = (cost * (3 / 2) * (gamma ** 2) * price / self.a) ** (1 / 3)

        return width



import torch
from torch import Tensor

from ..nn.functional import european_binary_payoff
from ._base import Derivative


class EuropeanBinaryOption(Derivative):
    """
    A European binary option.

    An American binary call option pays an unit amount of cash if and only if
    the underlying asset's price at maturity is equal or greater than the strike price.

    An American binary put option pays an unit amount of cash if and only if
    the underlying asset's price at maturity is equal or smaller than the strike price.

    The payoff of an American binary call option is given by:

    .. math::

        \\mathrm{payoff} =
        \\begin{cases}
            1 & (S \\geq K) \\\\
            0 & (\\text{otherwise})
        \\end{cases}

    with :math:`S` being the underlying asset's price at maturity and
    :math:`K` being the strike price (`strike`) of the option

    The payoff of an American binary put option is given by:

    .. math::

        \\mathrm{payoff} =
        \\begin{cases}
            1 & (S \\leq K) \\\\
            0 & (\\text{otherwise})
        \\end{cases}

    Args:
        underlier (:class:`Primary`): The underlying instrument of the option.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1): The strike price of the option.
        maturity (float, default=20/250) The maturity of the option.
        dtype (torch.dtype, optional): Desired device of returned tensor.
            Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see `torch.set_default_tensor_type()`).
            `device` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Examples:

        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanBinaryOption
        >>> _ = torch.manual_seed(42)
        >>> deriv = EuropeanBinaryOption(BrownianStock(), maturity=5/250)
        >>> deriv.simulate(n_paths=2)
        >>> deriv.underlier.prices
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930],
                [1.0000, 1.0282, 1.0199, 1.0258, 1.0292]])
        >>> deriv.payoff()
        tensor([0., 1.])
    """

    def __init__(
        self,
        underlier,
        call: bool = True,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        dtype: torch.dtype = None,
        device: torch.device = None,
    ):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")
        params += self.dinfo
        return self.__class__.__name__ + "(" + ", ".join(params) + ")"

    def payoff(self) -> Tensor:
        return european_binary_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



import torch

from ._base import Feature


class Moneyness(Feature):
    """Moneyness of the underlying instrument of the derivative.

    Args:
        log (bool, default=False): If `True`, represents log moneyness.
    """

    def __init__(self, log=False):
        super().__init__()

        self.log = log

    def __str__(self):
        if self.log:
            return "log_moneyness"
        else:
            return "moneyness"

    def __getitem__(self, i):
        # spot price: shape (N, 1)
        s = self.derivative.underlier.prices[..., i].reshape(-1, 1)
        output = s / self.derivative.strike
        if self.log:
            output = torch.log(output)
        return output


class LogMoneyness(Moneyness):
    """Log moneyness of the underlying instrument of the derivative."""

    def __init__(self):
        super().__init__(log=True)


class ExpiryTime(Feature):
    """Remaining time to the maturity of the derivative."""

    def __str__(self):
        return "expiry_time"

    def __getitem__(self, i):
        value = self.derivative.maturity - i * self.derivative.underlier.dt
        return torch.full_like(self.derivative.underlier.prices[:, :1], value)


class Volatility(Feature):
    """Volatility of the underlier of the derivative."""

    def __str__(self):
        return "volatility"

    def __getitem__(self, i):
        value = self.derivative.underlier.volatility
        return torch.full_like(self.derivative.underlier.prices[:, :1], value)


class PrevHedge(Feature):
    """Previous holding of underlier."""

    def __str__(self):
        return "prev_hedge"

    def __getitem__(self, i):
        if hasattr(self.hedger, "prev_output"):
            return self.hedger.prev_output.reshape(-1, 1)
        else:
            # prices: shape (N, T)
            return torch.zeros_like(self.derivative.underlier.prices[:, :1])


class Barrier(Feature):
    """A feature which signifies whether the price of the underlier have reached
    the barrier. The output `1.0` means that the price have touched the barrier,
    and `0` otherwise.

    Args:
        barrier (float): The price level of the barrier.
        up (bool, default True): If `True`, signifies whether the price has exceeded
            the barrier upward.
            If `False`, signifies whether the price has exceeded the barrier downward.
    """

    def __init__(self, barrier, up=True):
        super().__init__()

        self.barrier = barrier
        self.up = up

    def __repr__(self):
        return self.__class__.__name__ + f"({self.barrier}, up={self.up})"

    def __getitem__(self, i):
        if self.up:
            # shape: (N, i)
            touch_barrier = (
                self.derivative.underlier.prices[..., : i + 1] >= self.barrier
            )
        else:
            touch_barrier = (
                self.derivative.underlier.prices[..., : i + 1] <= self.barrier
            )
        return touch_barrier.any(dim=-1, keepdim=True).to(
            self.derivative.underlier.prices
        )


class Zero(Feature):
    """A feature of which value is always zero."""

    def __str__(self):
        return "zero"

    def __getitem__(self, i):
        return torch.zeros_like(self.derivative.underlier.prices[:, :1])


class MaxMoneyness(Feature):
    """Cumulative maximum of moneyness.

    Args:
        log (bool, default=False): If `True`, represents log moneyness.
    """

    def __init__(self, log=False):
        self.log = log
        self.moneyness = Moneyness(log=log)

    def __str__(self):
        if self.log:
            return "max_log_moneyness"
        else:
            return "max_moneyness"

    def __getitem__(self, i):
        s = (
            self.derivative.underlier.prices[..., : i + 1]
            .max(dim=1, keepdim=True)
            .values
        )
        output = s / self.derivative.strike
        if self.log:
            output = torch.log(output)
        return output

    def of(self, derivative=None, hedger=None):
        super().of(derivative)
        self.moneyness.of(derivative=derivative, hedger=hedger)
        return self


class MaxLogMoneyness(MaxMoneyness):
    """Cumulative maximum of log Moneyness."""

    def __init__(self):
        super().__init__(log=True)


class ModuleOutput(Feature, torch.nn.Module):
    """The feature computed as an output of a `torch.nn.Module`.

    Args:
        module (torch.nn.Module): Module to compute the value of the feature.
            The input and output shapes should be `(N, *, H_in) -> (N, *, 1)`,
            where `N` stands for the number of Monte Carlo paths of the underlier of
            the derivative, `H_in` stands for the number of input features
            (namely, `H_in = len(inputs)`),
            and `*` means any number of additional dimensions.
        inputs (list[Feature]): The input features to the module.

    Examples:

        >>> from torch.nn import Linear
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> deriv = EuropeanOption(BrownianStock())
        >>> deriv.simulate(n_paths=3)
        >>> m = Linear(2, 1)
        >>> f = ModuleOutput(m, [Moneyness(), ExpiryTime()]).of(deriv)
        >>> f[0]
        tensor([[...],
                [...],
                [...]], grad_fn=<AddmmBackward>)
        >>> f
        ModuleOutput(
          inputs=['moneyness', 'expiry_time'],
          (module): Linear(in_features=2, out_features=1, bias=True)
        )

        >>> from pfhedge.nn import BlackScholes

        >>> _ = torch.manual_seed(42)
        >>> deriv = EuropeanOption(BrownianStock())
        >>> deriv.simulate(n_paths=3)
        >>> m = BlackScholes(deriv)
        >>> f = ModuleOutput(m, [LogMoneyness(), ExpiryTime(), Volatility()]).of(deriv)
        >>> f[0]
        tensor([[...],
                [...],
                [...]])
    """

    def __init__(self, module, inputs):
        super().__init__()

        self.module = module
        self.inputs = inputs

    def extra_repr(self):
        return f"inputs={[str(f) for f in self.inputs]},"

    def forward(self, input):
        return self.module(input)

    def __getitem__(self, i):
        return self(torch.cat([f[i] for f in self.inputs], 1))

    def of(self, derivative=None, hedger=None):
        super().of(derivative, hedger)
        self.inputs = [feature.of(derivative, hedger) for feature in self.inputs]
        return self



from ._base import Derivative
from ._base import Primary
from .american_binary import AmericanBinaryOption
from .european import EuropeanOption
from .european_binary import EuropeanBinaryOption
from .lookback import LookbackOption
from .underlier import BrownianStock
